"""Top-level package for SCRUB."""

__author__ = """Lyle Barner"""
__email__ = 'lyle.barner@jpl.nasa.gov'
__version__ = '2.6.4'

# Check the Python version for compatibility
import sys
if not sys.version_info >= (3, 6):
    print('ERROR: Python version 3.6 or greater is required to run SCRUB')
    sys.exit(100)



import re
import pathlib
import logging
from scrub.tools.parsers import translate_results


# Initialize variables
suppression_lines = []


def micro_filter_check(source_file, warning_line, valid_warning_types):
    """This function checks to see if a warning has been marked as a false positive by the user.

    Inputs:
        - source_file: Absolute path to the source code file of interest [string]
        - warning_line: Line of interest of source code file [int]
        - valid_warning_types: List of strings containing valid types for each tool [list of strings]

    Outputs:
        - ignore_line: Indicator if warning should be ignored [bool]
    """
    # Set the base string
    ignore_base = "scrub_ignore_warning"

    # Initialize the return value
    ignore_line = False

    try:
        # Get the line of interest
        with open(source_file, 'r', errors='ignore') as input_fh:
            # Read the lines
            if warning_line == 0:
                line = input_fh.readlines()[0]
            else:
                line = input_fh.readlines()[warning_line - 1]

        # Check for suppression syntax
        if (ignore_base in line.lower()) or ('@suppress' in line.lower()):
            for check_type in valid_warning_types:
                if check_type in line.lower():
                    # Print a status message
                    logging.debug('\tWarning removed - Warning has been marked as a false positive')
                    logging.debug('\t\t%s', line)

                    # Update the output
                    ignore_line = True

                    break

    except IOError:
        logging.warning('\t\tMicro-filter warning')
        logging.warning('\t\tFile %s could not be found.', source_file)

    # Return value
    return ignore_line


def ignore_query_check(warning_tool, warning_query, ignore_queries_file):
    """This function checks if a result should be skipped based on the type of query.

    Inputs:
        - line: Line of interest from SCRUB results [string]
        - ignore_queries_file: Full path to the SCRUBExcludeQueries file [string]

    Outputs:
        - skip: Indicator if result should be filtered out [bool]
    """

    # Initialize the variables
    skip = False

    # Import the ignore data
    if ignore_queries_file.is_file():
        with open(ignore_queries_file, 'r') as ignore_fh:
            ignore_queries = ignore_fh.readlines()

        # Iterate through every line of the ignore data
        for ignore_line in ignore_queries:
            # Split the line and store the values
            ignore_line_split = list(filter(None, re.split(':', ignore_line.strip())))
            ignore_tool = ignore_line_split[0].strip().lower()
            ignore_query = ignore_line_split[1].strip()

            # Determine if the ine should be skipped
            if (ignore_tool == warning_tool) and (ignore_query == warning_query):
                skip = True

                # Print a status message
                logging.debug('\tWarning removed - Warning generated by a filtered query')
                logging.debug('\t\t%s: %s', warning_tool, warning_query)

    return skip


def external_warning_check(warning_file, source_root):
    """This function checks to see if a warning originates outside the source code directory.

    Inputs:
        - warning_file:
        - source_root: Full path to the top level directory of the source code [string]

    Outputs:
        - skip: Indicator if result should be filtered out [bool]
    """

    # Check to see if the file exists outside the source root
    if source_root not in warning_file.parents:
        skip = True

        # Print a status message
        logging.debug('\tWarning removed - Warning occurs in a file that is outside the source root')
        logging.debug('\t\t%s', warning_file)
    else:
        skip = False

    return skip


def baseline_filtering_check(warning_file, excluded_files):
    """This function checks to see if a warning occurs in a directory or file that should be ignored.

    Inputs:
        - warning_file: File of interest from warning data [string]
        - excluded_files: Set of files read from SCRUBAnalysisFilteringList file [set [string]]

    Outputs:
        - skip: Indicator if result should be filtered out [bool]
    """
    # Iterate through every line of the ignore data
    if warning_file in excluded_files:
        logging.debug('\tWarning removed - Warning occurs in a file that has been excluded from analysis')
        logging.debug('\t\t%s', warning_file)
        return True
    return False


def check_filtering_file(filtering_file, create):
    """This function checks the filtering file to see if it exists.

    Inputs:
        - filtering_file: Full path to the filtering file to be checked [string]
        - create: Flag to indicate if the file should be created if it doesn't exist [bool]
    """

    # Check to see if all the files exist
    if not filtering_file.is_file():
        logging.info('No %s file exists.', str(filtering_file))

        # Create the file if necessary
        if create:
            logging.info('\tCreating blank filtering file %s.', str(filtering_file))
            logging.info('\tPlease add filtering patterns to this file.')
            open(filtering_file, 'w+').close()


def duplicate_check(warning, warning_log):
    """This function checks to make sure that the warning has not been reported before.

    Inputs:
        - warning: current warning to be checked [string]
        - warning_log: log containing all warnings written previously [list of strings]

    Outputs:
        - skip: Indicator if result should be filtered out [bool]
    """

    # Initialize the variables
    skip = False

    # Check if the waring has been written out before
    if warning in warning_log:
        skip = True

    return skip


def filter_results(warning_list, output_file, filtering_file, ignore_query_file, source_root, enable_micro_filtering,
                   enable_external_warnings, valid_warning_types):
    """This function performs the filtering, including all other filtering functions.

    Inputs:
        - input_files: List of absolute paths to the input file(s) of interest [list of string]
        - output_file: Absolute path to file where filtered results will be stored [string]
        - filtering_file: Absolute path to the SCRUBAnalayisFilteringList file [string]
        - ignore_query_file: Absolute path to the SCRUBExcludeQueries file [string]
        - source_root: Absolute path to the top level directory of the source code [string]
        - enable_micro_filtering: Flag to enable/disable micro filtering [logical]
        - enable_external_warnings: Flag to enable/disable external warnings [logical]
        - valid_warning_types: List that contains valid warning type tags [list of strings]

    Outputs:
        - output_file: All filtered results are written to the output_file
    """

    # Initialize the variables
    filtered_warnings = []
    if output_file.stem == 'p10':
        valid_warning_types.append('p10')

    # Import the ignore data
    with open(filtering_file, 'r') as input_fh:
        excluded_files = set([x.strip() for x in input_fh.readlines()])

    # Print a log message
    logging.info('')
    logging.info('\tFiltering results...')
    logging.info('\t>> Executing command: filter_results.filter_results(<warning_list>, %s, %s, %s, %s, %r, %r)',
                 output_file, filtering_file, ignore_query_file, source_root, enable_micro_filtering,
                 enable_external_warnings)
    logging.info('\t>> From directory: %s', str(pathlib.Path().absolute()))

    # Update the source root to make it absolute
    source_root = source_root.resolve()

    # Iterate through every warning in the list
    for warning in warning_list:
        # Check to see if it should be ignored
        if baseline_filtering_check(warning['file'], excluded_files):
            continue

        # Check to see if the warning is external to the source directory
        if not enable_external_warnings \
                and external_warning_check(warning['file'], source_root):
            continue

        # Perform micro filtering checking
        if enable_micro_filtering and micro_filter_check(warning['file'], warning['line'], valid_warning_types):
            continue

        # Check to see if the query should be ignore
        if ignore_query_check(warning['tool'], warning['query'], ignore_query_file):
            continue

        # If we made it here we want the warning.
        # Make the warning file path relative and append to filtered list
        warning['file'] = warning['file'].relative_to(source_root)
        filtered_warnings.append(warning)

    # Write out the results
    logging.info('\t>> Results filtered. Writing {}.'.format(output_file))

    translate_results.create_scrub_output_file(filtered_warnings, output_file)



import shutil


def clean_subdirs(root_dir):
    """This function cleans all sub-directories under the specified root directory.

    Inputs:
        - root_dir: Absolute path to the root directory of interest [string]
    """

    # Find all the files and directories of interest
    for scrub_dir in list(root_dir.rglob('.scrub')):
        if scrub_dir != root_dir.joinpath('.scrub'):
            shutil.rmtree(scrub_dir)


def clean_directory(directory):
    """This function removes previous SCRUB data products.

    Inputs:
        - directory: Full path to the top-level source code directory [string]
    """

    # Print a status message
    print('        Removing previous SCRUB results from source tree...\n')

    # Remove the root directory
    if directory.joinpath('.scrub').exists():
        shutil.rmtree(directory.joinpath('.scrub'))

    # Remove all of the sub-directories
    clean_subdirs(directory)



import re
import os
import pathlib
import logging


def parse_filtering_file(file_path):
    """This function parses a regex filtering file to obtain filtering options for SCRUB.

    Inputs:
        - file_path: Absolute path to regex filtering file [string]

    Outputs:
        - filtering_options: List of filtering options to be used [list of strings]
    """

    # Initialize variables
    filtering_options = []

    # Read in the default filtering options
    with open(file_path, 'r') as input_fh:
        # Iterate through every line of the file
        for line in input_fh.readlines():

            # Check to make sure the line is formatted correctly
            if len(line.strip()) > 2 and line.startswith(('-', '+')):
                # Append the filtering option
                if line[1] == ' ':
                    filtering_options.append((line[0], line.strip()[2:]))
            elif not line.startswith('#'):
                logging.warning('\tInvalid regex filtering line: {}'.format(line.strip()))

    return filtering_options


def create_file_list(source_root_dir, filtering_output_file, filtering_options_file, initial_filtering_list=''):
    """This function creates a list of the files that will be included in SCRUB analysis.

    Inputs:
        - source_root_dir: Absolute path to the source code root directory [string]
        - filtering_output_file: Absolute path to the output file containing a list of files to be included in analysis
                                 [string]
        - filtering_options_file: Absolute path to the filtering file to be used to create list of analysis files
                                  [string]
        - initial_filtering_list: Absolute path to the file containing an initial set of files [string]
    """

    # Print a status message
    logging.info('')
    logging.info('\tCreating analysis filtering list...')
    logging.info('\t>> Executing command: create_file_list.create_file_list(%s, %s, %s, %s)', source_root_dir,
                 filtering_output_file, filtering_options_file, initial_filtering_list)
    logging.info('\t>> From directory: %s', pathlib.Path.cwd())

    # Initialize the variables
    raw_file_list = []
    default_filtering_options_file = pathlib.Path(__file__).parent.joinpath('FilteringDefaults').resolve()

    # Create the list of all available files, if necessary
    if initial_filtering_list:
        # Read in the list
        with open(initial_filtering_list, 'r') as input_fh:
            raw_file_list = input_fh.readlines()

    else:
        for root, dir_names, file_names, in os.walk(source_root_dir, topdown=True):
            dir_names[:] = [d for d in dir_names if d not in ['.scrub']]
            for file_name in file_names:
                raw_file_list.append(os.path.join(root, file_name))

    # Read in the values from the filtering file and add them to the list
    if filtering_options_file.exists():
        # Print a status message
        logging.info('')
        logging.info('\tParsing input file.')

        # Parse the filtering file
        filtering_options = (parse_filtering_file(filtering_options_file) +
                             parse_filtering_file(default_filtering_options_file))

    else:
        # Print a warning message
        logging.info('')
        logging.info('\tNo filtering file was found or no filtering file was provided. Using default values.')

        # Parse the filtering file
        filtering_options = parse_filtering_file(default_filtering_options_file)

    # Modify the list based on the include and exclude options
    filtered_file_list = raw_file_list.copy()
    for filtering_option in filtering_options:
        for file_path in list(filter(re.compile(filtering_option[1]).search, raw_file_list)):
            if filtering_option[0] == '-':
                filtered_file_list.remove(file_path)

            elif filtering_option[0] == '+':
                filtered_file_list.append(file_path)

    # Print the results to the output file
    filtered_file_list.sort()
    with open(filtering_output_file, 'w') as output_fh:
        for filtered_file in filtered_file_list:
            if pathlib.Path(filtered_file).anchor == '/':
                relative_path = pathlib.Path(filtered_file).relative_to(source_root_dir)
            else:
                relative_path = pathlib.Path(filtered_file)
            output_fh.write('%s\n' % relative_path)



import sys
import json
import pathlib
from scrub.tools.parsers import translate_results

WARNING_LEVEL = 'Low'
ID_PREFIX = 'coverity'


def parse_json(raw_input_file, parsed_output_file):
    """This function parses the Coverity internal JSON results format into SCRUB formatted results.

    Inputs:
        - raw_input_file: Absolute path to the file containing raw Coverity warnings [string]
        - parsed_output_file: Absolute path to the file where the parsed warnings will be stored [string]
    """

    # Initialize variables
    coverity_issues = []
    warning_count = 1

    # Read in the input file
    with open(raw_input_file, 'r') as input_fh:
        input_data = json.load(input_fh)

    # Iterate through every issue
    for issue in input_data['issues']:
        # Parse issue data
        warning_id = '%s%03d' % (ID_PREFIX, warning_count)
        warning_file = pathlib.Path(issue['mainEventFilePathname'])
        warning_line = int(issue['mainEventLineNumber'])
        warning_checker = issue['checkerName']
        warning_description = []

        # Get the warning description
        for event in issue['events']:
            if event['eventTag'] != 'caretline':
                warning_description.append('%s:%s:' % (event['strippedFilePathname'], event['lineNumber']))
                warning_description.append('%s: %s' % (event['eventTag'], event['eventDescription']))

        coverity_issues.append(translate_results.create_warning(warning_id, warning_file, warning_line,
                                                                warning_description, 'coverity', WARNING_LEVEL,
                                                                warning_checker))

        # Increment the warning count
        warning_count = warning_count + 1

    # Create the output file
    translate_results.create_scrub_output_file(coverity_issues, parsed_output_file)


if __name__ == '__main__':
    parse_json(pathlib.Path(sys.argv[1]), pathlib.Path(sys.argv[2]))



import sys
import json
import pathlib
from scrub.tools.parsers import translate_results

WARNING_LEVEL = 'Low'
ID_PREFIX = 'sonarqube'


def parse_warnings(results_dir, parsed_output_file, source_root, sonarqube_url):
    """This function parses the raw SonarQube warnings into the SCRUB format.
        Inputs:
            - results_dir: Absolute path to the raw SonarQube output file directory [string]
            - parsed_output_file: Absolute path to the file where the parsed warnings will be stored [string]
            - source_root: Absolute path to the source root directory [string]
            - sonarqube_url: URL of the SonarQube server [string]
        """

    # Initialize the variables
    warning_count = 1
    raw_warnings = []

    # Find all of the hotspot files
    hotspot_files = results_dir.glob('sonarqube_hotspot_[0-9]*.json')

    # Iterate through every hotspot file
    for hotspot_file in hotspot_files:
        # Read in the input data
        with open(hotspot_file, 'r', encoding='utf-8') as input_fh:
            hotspot_data = json.loads(input_fh.read())

        # Parse the finding
        warning_file = source_root.joinpath(hotspot_data['component']['path'])
        warning_message = hotspot_data['rule']['name'].splitlines()
        warning_query = hotspot_data['rule']['key']
        warning_id = ID_PREFIX + str(warning_count).zfill(3)
        warning_link = sonarqube_url + '/security_hotspots?id=' + hotspot_data['project']['key'] + '&hotspots=' + hotspot_data['key']

        # Get the line
        if 'line' in hotspot_data.keys():
            warning_line = int(hotspot_data['line'])
        else:
            warning_line = 0

        # Check to see if the warning should be suppressed
        if 'resolution' in hotspot_data.keys():
            suppression = True
        else:
            suppression = False

        # Add the link to the message
        warning_message.append(warning_link)

        # Add to the warning dictionary
        raw_warnings.append(translate_results.create_warning(warning_id, warning_file, warning_line,
                                                             warning_message, ID_PREFIX, WARNING_LEVEL,
                                                             warning_query, suppression))

        # Increment the warning count
        warning_count = warning_count + 1

    # Find all of the issues files
    issues_files = results_dir.glob('sonarqube_issues_[0-9]*.json')

    # Iterate through every issues results file
    for issues_file in issues_files:
        # Read in the input data
        with open(issues_file, 'r', encoding='utf-8') as input_fh:
            issues_data = json.loads(input_fh.read())

        # Iterate through every finding in the input file
        for issue in issues_data['issues']:
            # Check to see if the warning should be suppressed
            if 'resolution' in issue.keys():
                suppression = True
            else:
                suppression = False

            # Parse the finding
            warning_file = source_root.joinpath(issue['component'].split(':')[-1]).resolve()
            warning_message = issue['message'].splitlines()
            warning_id = ID_PREFIX + str(warning_count).zfill(3)

            # Get a link to the finding
            warning_link = sonarqube_url + '/project/issues?id=' + issue['project'] + '&open=' + issue['key']

            # Add the link to the warning message
            warning_message.append(warning_link)

            # Parse the query if it exists
            if 'rule' in issue.keys():
                warning_query = issue['rule']
            else:
                warning_query = ''

            # Get the line number
            if 'textRange' in issue.keys():
                warning_line = int(issue['textRange']['startLine'])
            else:
                warning_line = 0

            # Add to the warning dictionary
            raw_warnings.append(translate_results.create_warning(warning_id, warning_file, warning_line,
                                                                 warning_message, ID_PREFIX, WARNING_LEVEL,
                                                                 warning_query, suppression))

            # Increment the warning count
            warning_count = warning_count + 1

    # Create the SCRUB output file
    translate_results.create_scrub_output_file(raw_warnings, parsed_output_file)


if __name__ == '__main__':
    parse_warnings(pathlib.Path(sys.argv[1]), pathlib.Path(sys.argv[2]), pathlib.Path(sys.argv[3]), sys.argv[4])



import re
import pathlib
import logging
import traceback
from scrub.utils import scrub_utilities


def distribute_warnings(warning_file, source_dir):
    """This function moves warnings to be co-located with the source file of interest.

    Inputs:
        - warning_file: Full path to the file containing SCRUB-formatted warnings [string]
        - source_dir: Full path to the top-level directory of the source code [string]

    Outputs:
        - A series of .scrub directories and output files will be created as necessary
    """

    # Initialize the variables
    warning_type = warning_file.stem

    # Print a status message
    logging.info('')
    logging.info('\tMoving results...')
    logging.info('\t>> Executing command: do_gui.distribute_warnings(%s, %s)', str(warning_file), str(source_dir))
    logging.info('\t>> From directory: %s', str(pathlib.Path().absolute()))

    # Import the warning file
    with open(warning_file, 'r') as input_fh:
        input_data = input_fh.readlines()

    # Update the source root to make it absolute
    source_dir = source_dir.resolve()

    # Iterate through every line of the file
    for i in range(0, len(input_data)):
        # Set the line
        line = input_data[i]

        # Check to see if the line contains a warning
        if re.search(r'<.*>.*:.*:.*:', line):
            # Add the line to the warning text
            warning = line

            # Get the warning file
            warning_file = pathlib.Path(line.split(":")[1].strip())
            warning_file_absolute = source_dir.joinpath(warning_file)

            # Make sure the warning file is within the source root, but not at the source root
            if warning_file_absolute.exists() and (source_dir != warning_file_absolute.parent):
                # Get the rest of the warning text
                j = i + 1
                while input_data[j].strip() != '':
                    warning = warning + input_data[j]

                    # Increment the line
                    j = j + 1

                # Get the warning directory
                warning_directory = warning_file_absolute.parent

                # Create the scrub output paths
                local_scrub_directory = warning_directory.joinpath('.scrub')
                local_scrub_warning_file = local_scrub_directory.joinpath(warning_type)

                # Create a .scrub directory if it doesn't already exists
                if not local_scrub_directory.exists():
                    local_scrub_directory.mkdir()
                    local_scrub_directory.chmod(0o755)

                # Write the warning to the output file
                with open(local_scrub_warning_file, 'a') as output_fh:
                    output_fh.write('%s\n' % warning_file.name)

                # Change the permissions of the output file
                local_scrub_warning_file.chmod(0o644)


def initialize_analysis(tool_conf_data):
    """The purpose of this function is to prepare the tool to perform analysis.

    Inputs:
        - tool_conf_data: Dictionary of scrub.cfg input variables [dict]
    """

    # Initialize the derived variables
    gui_log_file = tool_conf_data.get('scrub_log_dir').joinpath('gui.log')

    # Add derived values to the dictionary
    tool_conf_data.update({'gui_log_file': gui_log_file})


def run_analysis(baseline_conf_data, console_logging=logging.INFO, override=False):
    """This function runs this module based on input from the scrub.cfg file

    Inputs:
        - baseline_conf_data: Dictionary of raw scrub.cfg configuration parameters [dict]
        - console_logging: Level of console logging information to print to console [optional] [enum]
        - override: Force tool execution? [optional] [bool]
    """

    # Import the config file data
    tool_conf_data = baseline_conf_data.copy()
    initialize_analysis(tool_conf_data)

    # Initialize variables
    gui_exit_code = 2
    attempt_analysis = tool_conf_data.get('scrub_gui_export') or override

    # Determine if GUI export can be run
    if attempt_analysis:
        try:
            # Create the logging file, if it doesn't already exist
            scrub_utilities.create_logger(tool_conf_data.get('gui_log_file'), console_logging)

            # Print a status message
            logging.info('')
            logging.info('Perform GUI export...')

            # Get a list of the filtered SCRUB output files
            filtered_output_files = tool_conf_data.get('scrub_analysis_dir').glob('*.scrub')

            # Move the warnings to the appropriate directories
            for filtered_output_file in filtered_output_files:
                distribute_warnings(filtered_output_file, tool_conf_data.get('source_dir'))

            # Set the exit code
            gui_exit_code = 0

        except:     # lgtm [py/catch-base-exception]
            # Print a warning message
            logging.warning('GUI export could not be performed. Please see log file %s for more information.',
                            str(tool_conf_data.get('gui_log_file')))

            # Print the exception traceback
            logging.warning(traceback.format_exc())

            # Set the exit code
            gui_exit_code = 1

        finally:
            # Close the loggers
            logging.getLogger().handlers = []

            # Update the permissions of the log file if it exists
            if tool_conf_data.get('gui_log_file').exists():
                tool_conf_data.get('gui_log_file').chmod(0o644)

    # Return the exit code
    return gui_exit_code



import re
import os
import sys
import time
import json
import shutil
import pathlib
import logging
import threading
import subprocess
import configparser
import argparse
import urllib.request
from scrub.tools.parsers import translate_results


class Spinner:
    busy = False
    delay = 0.1

    @staticmethod
    def spinning_cursor():
        while 1:
            for cursor in '|/-\\':
                yield cursor

    def __init__(self, delay=None):
        self.spinner_generator = self.spinning_cursor()
        if delay and float(delay):
            self.delay = delay

    def spinner_task(self):
        while self.busy:
            sys.stdout.write(next(self.spinner_generator))
            sys.stdout.flush()
            time.sleep(self.delay)
            sys.stdout.write('\b')
            sys.stdout.flush()

    def __enter__(self):
        self.busy = True
        threading.Thread(target=self.spinner_task).start()

    def __exit__(self, exception, value, tb):
        self.busy = False
        time.sleep(self.delay)
        if exception is not None:
            return False


def get_pip_version():
    """This function gets the latest available version number from pip.

    Inputs:
        - None

    Outputs:
        - version_number: Latest pip version that is available [string]
    """

    # Get the version number
    try:
        api_data = json.loads(urllib.request.urlopen('https://pypi.org/pypi/nasa-scrub/json').read())
        version_number = api_data['info']['version']
    except:
        version_number = 'Unknown'

    return version_number


def parse_template(template_file, output_file, conf_data):
    """This function parsers analysis templates and populates them with configuration values.

    Inputs:
        - template_file: Absolute path to analysis template file [Path object]
        - output_file: Absolute path to output analysis script [Path object]
        - conf_data: Dictionary of values read from configuration file [dict]
    """

    # Read in the analysis template
    with open(template_file, 'r') as input_fh:
        template_data = input_fh.read()

    # Replace all of the variables with config data
    for key in conf_data.keys():
        if isinstance(conf_data.get(key), bool):
            template_data = template_data.replace('${{' + key.upper() + '}}', str(conf_data.get(key)).lower())
        else:
            template_data = template_data.replace('${{' + key.upper() + '}}', str(conf_data.get(key)))

    # Write out the completed template
    with open(output_file, 'w') as output_fh:
        output_file.chmod(0o777)
        output_fh.write('%s' % template_data)

    # Check the contents of the analysis script file
    check_artifact(output_file, True)


class CommandExecutionError(Exception):
    pass


def check_log_file(log_file):
    """This function checks log files for potential issues in analysis log files

    Inputs:
        - log_file: Absolute path to the log file of interest [string]

    Outputs:
        - errors: Were errors found? [boolean]
    """

    # Initialize variables
    errors = False

    # Check the size of the log file
    if log_file.stat().st_size == 0:
        errors = True

    # Import the log file
    with open(log_file, 'r') as input_fh:
        for line in input_fh:
            if 'ERROR:' in line:
                # Update return value
                errors = True
                break

    return errors


def check_artifact(input_artifact, critical=False):
    """This function checks to ensure the given file is not empty.

    Inputs:
        - input_artifact: Absolute path to the artifact of interest [string]
        - critical: Indication of artifact criticality [bool]
            - True: Artifact is critical and should not be empty
    """

    # Get the size of the item
    if input_artifact.is_file():
        size = input_artifact.stat().st_size
    else:
        size = len(list(input_artifact.iterdir()))

    # Check to make sure the file isn't empty
    if size == 0:
        if critical:
            message = input_artifact.name + ' is empty. This should not be empty.'
            raise CommandExecutionError(message)
        else:
            logging.warning('')
            logging.warning('\t%s is empty.', str(input_artifact.name))
            logging.warning('\tThis may or may not be a problem.')


def split_results(baseline_file, subset_file, remainder_file, queries):
    """This function splits a baseline set of SCRUB results into two separate files based on a given query list.

    Inputs:
        - baseline_file: Absolute path to the baseline file containing SCRUB warnings [string]
        - subset_file: Absolute path to the output file to contain filtered warnings [string]
        - remainder_file: Absolute path to the output file to contain remaining warnings not filtered [string]
        - query_list: List of query strings to filter on [list of strings]
    """

    # Import the contents of the baseline input file
    with open(baseline_file, 'r') as input_fh:
        baseline_data = input_fh.readlines()

    try:
        # Open the output files
        subset_fh = open(subset_file, 'w')
        remainder_fh = open(remainder_file, 'w')

        # Iterate through every line of the baseline data
        for i in range(0, len(baseline_data)):
            line = baseline_data[i]

            # Check to see if the line contains a warning header
            if re.search(translate_results.WARNING_LINE_REGEX, line):
                # Get the current warning query
                warning_query = list(filter(None, re.split(":", line)))[-1].strip()

                # Get the full warning content
                current_warning = get_warning(baseline_data, i)

                # Check to see if the query is in the queries list of interest
                if warning_query in queries:
                    # Write the query to the output file
                    subset_fh.write('%s' % current_warning)

                else:
                    # If not, write the result to to remaining output file
                    remainder_fh.write('%s' % current_warning)

    finally:
        # Close the files
        subset_fh.close()
        remainder_fh.close()


def get_warning(warning_data, index):
    """This function gets a complete warning when passed the index of a warning header.

    Inputs:
        - warning_data: List of lines of warning file contents [list of strings]
        - index: Index of warning header of interest [int]

    Outputs:
        - warning_content: Full warnings contents including header data [string]
    """

    # Initialize variables
    warning_content = warning_data[index]

    # Get the warning content
    for i in range(index + 1, len(warning_data)):
        # Get the line content
        line = warning_data[i]

        # Make sure you haven't hit the next warning header
        if re.search(translate_results.WARNING_LINE_REGEX, line):
            break
        else:
            warning_content = warning_content + line

    return warning_content


def execute_command(call_string, my_env, output_file=None, interactive=False):
    """This function executes a command string and captures the results.

    Inputs:
        - call_string: Command to execute in the shell [string]
        - my_env: Environment to use during execution [dict]
        - output_file: Absolute path to output file for storing results [string] [optional]
        - interactive: Open command for user input? [bool] [optional]
    """

    # Initialize variables
    output_data = ''

    # Write out a logging message
    logging.info('')
    logging.info('    >> Executing command: %s', call_string)
    logging.info('    >> From directory: %s', str(pathlib.Path().absolute()))
    logging.debug('    Console output:')

    # Execute the call string and capture the output
    with Spinner():
        if interactive:
            proc = subprocess.Popen(call_string, shell=True, env=my_env, encoding='utf-8')
        else:
            proc = subprocess.Popen(call_string, shell=True, env=my_env, stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT, encoding='utf-8')

            # Write the output to the logging file
            for stdout_line in iter(proc.stdout.readline, ''):
                logging.debug('        %s', stdout_line.replace('\n', ''))
                output_data = output_data + stdout_line

            # Write results to the output file
            if output_file is not None:
                with open(output_file, 'w') as output_fh:
                    output_fh.write(output_data)

    # Wait for the process to finish
    proc.wait(timeout=None)

    # Throw an exception if necessary
    if proc.poll() > 0:
        raise CommandExecutionError


def create_logger(log_file, console_logging=logging.INFO):
    """This function creates the logger to be used for logging SCRUB data.

    Inputs:
        - log_file: Absolute path to the location of the log file to be created [string]
        - console_logging: Should debugging info be printed to the console?
            - Default value: logging.INFO
    """

    # Clear any existing loggers
    logging.getLogger().handlers = []

    # Create the logger, if it doesn't already exist
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)-8s %(message)s',
                        filename=str(log_file),
                        filemode='w')

    # Start the console logger
    console = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s %(levelname)-8s %(message)s')
    console.setFormatter(formatter)
    logging.getLogger('').addHandler(console)
    console.setLevel(console_logging)


def create_conf_file(output_path=None):
    """
    This function generates a blank configuration file at the desired output location.

    Inputs:
        --output: Path to desired output location [string] [optional]
            Default value: ./scrub_template.cfg
    """

    # Parse arguments if necessary
    if output_path is None:
        # Create the parser
        parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                         description=create_conf_file.__doc__)

        # Add parser arguments
        parser.add_argument('--output', default='./scrub_template.cfg')

        # Parse the arguments
        args = vars(parser.parse_args(sys.argv[2:]))
        output_path = args['output']

    # Initialize variables
    default_config_file = pathlib.Path(__file__).parent.joinpath('scrub_defaults.cfg').resolve()

    # Copy the default configuration file
    shutil.copyfile(default_config_file, output_path)


def parse_common_configs(user_conf_file, scrub_keys=[]):
    """This function parses a SCRUB configuration file and adds default values.

    Inputs:
        - conf_file: Absolute path to the SCRUB configuration file [string]
        - scrub_keys: List of configuration file sections to be retrieved [list of strings]

    Outputs:
        - scrub_conf_data: Dictionary of values read from configuration file [dict]
    """

    # Initialize the variables
    scrub_conf_data = {}
    scrub_init_path = pathlib.Path(__file__).parent.joinpath('scrub_defaults.cfg')

    # Read in the default config data
    scrub_init_data = configparser.ConfigParser()
    scrub_init_data.read(scrub_init_path)

    # Set the keys, if necessary
    if not scrub_keys:
        scrub_keys = scrub_init_data.sections()

    # Convert to a dictionary
    for key in scrub_keys:
        scrub_conf_data.update(dict(scrub_init_data.items(key)))

    # Read in the values from the conf file
    user_conf_data = configparser.ConfigParser()
    user_conf_data.read(user_conf_file)

    # Update the dictionary
    for user_section in user_conf_data.sections():
        for section_key in user_conf_data.options(user_section):
            # Update the value if the user conf has something
            if user_conf_data.get(user_section, section_key):
                scrub_conf_data.update({section_key: user_conf_data.get(user_section, section_key)})
            elif section_key not in scrub_conf_data.keys():
                # Add the key if it doesn't exist
                scrub_conf_data.update({section_key: user_conf_data.get(user_section, section_key)})

    # Update the configuration data
    for key in scrub_conf_data.keys():
        scrub_conf_data.update({key: os.path.expandvars(scrub_conf_data.get(key))})

        # # Update boolean values
        if scrub_conf_data.get(key).lower() == 'true':
            scrub_conf_data.update({key: True})
        elif scrub_conf_data.get(key).lower() == 'false':
            scrub_conf_data.update({key: False})

    # Process the language data to a standard format
    source_langs = list(filter(None, scrub_conf_data.get('source_lang').replace(' ', '').split(',')))
    for i, source_lang in enumerate(source_langs):
        if source_lang == 'c':
            source_langs[i] = 'cpp'
        elif source_lang == 'j':
            source_langs[i] = 'java'
        elif source_lang == 'p':
            source_langs[i] = 'python'
        elif source_lang == 'js':
            source_langs[i] = 'javascript'
    scrub_conf_data.update({'source_lang': ','.join(source_langs)})

    # Make the source root absolute
    scrub_conf_data.update({'source_dir': pathlib.Path(scrub_conf_data.get('source_dir')).expanduser().resolve()})

    # Make access tokens absolute
    scrub_conf_data.update({'codesonar_cert': pathlib.Path(scrub_conf_data.get('codesonar_cert')).resolve()})
    scrub_conf_data.update({'codesonar_key': pathlib.Path(scrub_conf_data.get('codesonar_key')).resolve()})

    # Set the SCRUB analysis directory
    scrub_conf_data.update({'scrub_analysis_dir': scrub_conf_data.get('source_dir').joinpath('.scrub')})

    # Set the default filtering file locations
    if scrub_conf_data.get('analysis_filters') == '':
        analysis_filters_file = user_conf_file.parent.joinpath('SCRUBFilters')
        scrub_conf_data.update({'analysis_filters': analysis_filters_file})
    if scrub_conf_data.get('query_filters') == '':
        analysis_filters_file = user_conf_file.parent.joinpath('SCRUBExcludeQueries')
        scrub_conf_data.update({'query_filters': analysis_filters_file})
    if scrub_conf_data.get('collaborator_filters') == '':
        collaborator_filters = user_conf_file.parent.joinpath('SCRUBCollaboratorFilters')
        scrub_conf_data.update({'collaborator_filters': collaborator_filters})

    # Set the SCRUB working directory
    if (scrub_conf_data.get('scrub_working_dir') is None) or (scrub_conf_data.get('scrub_working_dir') == ''):
        scrub_working_dir = scrub_conf_data.get('scrub_analysis_dir')
    else:
        scrub_working_dir = pathlib.Path(scrub_conf_data.get('scrub_working_dir')).expanduser().resolve()
    scrub_conf_data.update({'scrub_working_dir': scrub_working_dir})

    # Make every *path variable absolute and make every *build_dir variable absolute
    for key in scrub_conf_data.keys():
        if re.search(r'.+path', key) and (scrub_conf_data.get(key) != ''):
            path_value = scrub_conf_data.get(key)
            #path_value = os.path.abspath(os.path.expanduser(path_value))
            path_value = pathlib.Path(path_value).expanduser().resolve()
            scrub_conf_data.update({key: path_value})

        elif re.search(r'.+build_dir', key):
            if scrub_conf_data.get(key) == '':
                scrub_conf_data.update({key: scrub_conf_data.get('source_dir')})
            elif not scrub_conf_data.get(key).startswith(str(scrub_conf_data.get('source_dir'))):
                path_value = scrub_conf_data.get(key)
                #path_value = os.path.abspath(os.path.expanduser(path_value))
                path_value = pathlib.Path(path_value).expanduser().resolve()
                scrub_conf_data.update({key: path_value})

    # Add SCRUB root path
    scrub_path = pathlib.Path(__file__).parents[1]
    scrub_conf_data.update({'scrub_path': scrub_path})

    # Add the log directory
    scrub_log_dir = scrub_conf_data.get('scrub_analysis_dir').joinpath('log_files')
    scrub_conf_data.update({'scrub_log_dir': scrub_log_dir})

    # Add the raw results directory
    raw_results_dir = scrub_conf_data.get('scrub_analysis_dir').joinpath('raw_results')
    scrub_conf_data.update({'raw_results_dir': raw_results_dir})

    # Add the SARIF results directory
    sarif_results_dir = scrub_conf_data.get('scrub_analysis_dir').joinpath('sarif_results')
    scrub_conf_data.update({'sarif_results_dir': sarif_results_dir})

    # Add the filtering output file
    filtering_output_file = scrub_conf_data.get('scrub_analysis_dir').joinpath('SCRUBAnalysisFilteringList')
    scrub_conf_data.update({'filtering_output_file': filtering_output_file})

    return scrub_conf_data


def initialize_storage_dir(scrub_conf_data):
    """This function handles setting up the SCRUB analysis storage directory

    Inputs:
        - scrub_conf_data: Dictionary of values read from configuration file [dict]
    """

    # Create the .scrub analysis directory
    if not scrub_conf_data.get('scrub_analysis_dir').exists():
        scrub_conf_data.get('scrub_analysis_dir').mkdir()
        scrub_conf_data.get('scrub_analysis_dir').chmod(0o755)

    # Create the logging directory
    if not scrub_conf_data.get('scrub_log_dir').exists():
        scrub_conf_data.get('scrub_log_dir').mkdir()
        scrub_conf_data.get('scrub_log_dir').chmod(0o755)

    # Create the output directory
    if not scrub_conf_data.get('raw_results_dir').exists():
        scrub_conf_data.get('raw_results_dir').mkdir()
        scrub_conf_data.get('raw_results_dir').chmod(0o755)

    # Create the SARIF results directory
    if not scrub_conf_data.get('sarif_results_dir').exists():
        scrub_conf_data.get('sarif_results_dir').mkdir()
        scrub_conf_data.get('sarif_results_dir').chmod(0o755)

    # Create the analysis directory if it doesn't exist
    if scrub_conf_data.get('scrub_working_dir') != scrub_conf_data.get('scrub_analysis_dir'):
        if scrub_conf_data.get('scrub_working_dir').exists():
            print('ERROR: SCRUB storage directory ' + str(scrub_conf_data.get('scrub_working_dir')) +
                  ' already exists. Aborting analysis.')
            sys.exit(10)
        else:
            # Create the scrub working dir
            scrub_conf_data.get('scrub_working_dir').mkdir()
            scrub_conf_data.get('scrub_working_dir').chmod(0o755)
